#####################################################################
#
#   Penetrator - Text Indexing Module
#
#   Copyright (C) 2000/2003 Angel Ortega <angel@triptico.com>
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#   http://www.triptico.com
#
#####################################################################

package Penetrator;

use locale;

$VERSION = '3.1.6';

=pod

=head1 NAME

Penetrator - Text Indexing Engine

=head1 SYNOPSIS

 use Penetrator;

 # open a Penetrator object (using the DBM driver)
 $ph = new Penetrator("type" => "DBM",
		    "index_file" => "/var/index.db");
 # or

 # open a Penetrator object (using the DBI driver)
 $ph = new Penetrator("type" => "DBI",
		   "dbi_string" => $dbi_connection_string);

 # INDEXING

 # start indexing a document
 $ph->begin_doc_index($docname, $mtime);

 # feed the words
 open F, $docname;
 while(<F>)
 {
	$ph->doc_words(split(/\W+/,$_));
 }
 close F;

 # end indexing this document
 $ph->end_doc_index() or die "Doc couldn't be indexed";

 # SEARCHING

 # search all documents matching ALL The words
 $ph->search($word1, $word2);

 # fetch docnames
 while(my $d = $ph->fetch())
 {
	print "Words found in document '$d'\n";
 }

 # close connection
 $ph->close();

=head1 DESCRIPTION

Penetrator.pm is the engine of the Penetrator package. You can
retrieve it or get more documentation from its home page at

 http://www.triptico.com/software/penetrator.html

=head1 FUNCTIONS AND METHODS

=head2 B<new>

 $ph = new Penetrator("type"		  => $driver,
		  [ "min_word_size"	  => $size, ]
		  [ "zap"		  => $bool, ]
		  [ "normalize" 	  => \&function, ]
		  [ "disable_query_cache" => $bool, ]
		  [ %driver_specific_arguments ] );

Creates a new Penetrator object.

=over 4

=item I<type>

The object type references the driver to be used. It can be "DBM", to
use a Berkeley DB plain index, "GDBM", for a GNU dbm plain index,
"SDBM", for an SDBM plain index, "DBI", to use a relational database
using the Perl DBI interface, or "plain", for a dummy driver that
does not generate indexes, but searches every time. This argument is
mandatory.

=item I<min_word_size>

This is the minimal word size that will be taken into account
when building the database. By default is 0, meaning all words will
be indexed.

=item I<zap>

When this argument is set to a nonzero value, the database is
zapped, i.e., returned to an empty state. This can mean deleting
files or creating tables in an RDBMS.

=item I<normalize>

Sets the word normalization function for the database.
See L<ADVANCED TOPICS>.

=item I<disable_query_cache>

By default, successful queries are stored in a query cache to speed up
subsequent calls for the same information (useful, for example, when
navigating the results of a query from a CGI). If this argument is set
to zero, the query cache will not be used.

=back

=head2 B<begin_doc_index>

 $ph->begin_doc_index($docname, $mtime);

Starts indexing a document. I<$docname> is the name of the document;
though it used to be a file name, it can contain any textual and
unique reference. This is the key returned it search hits.
I<$mtime> is a numeric value containing the version of the document.
This usually contained the file modification time, but can contain
any other significant numeric value. If the document being indexed
already exists, the I<$mtime> parameter is tested, and the indexing
rejected if it's the same. Returns 1 if the document is accepted,
0 otherwise.

=head2 B<doc_words>

 $ph->doc_words($word [, $word ...] );

Add the word(s) to the document being indexed. The words cannot
contain spaces nor other separators (they are assumed to be the
result of a B<split(/\W+/)>).

=head2 B<end_doc_index>

 $ph->end_doc_index()

Ends the indexing of the document. Depending of the driver, this
can be the most time consuming function. It's return value must
be tested; if it's zero, the document was not indexed.

=head2 B<search>

 $ph->search($word [, $word... ]);

Searches the database for the documents matching ALL the words.
Words can also be prepended by a - or ~ to exclude documents
containing them from the final result. It does not return a value;
values must be retrieved using B<fetch()>.

=head2 B<fetch>

 $docname=$ph->fetch();

Returns a matching document, one at a time. The returned I<$docname>
is the first parameter of B<begin_doc_index()>. When no more hits
are left, undef is returned instead and the search is finished.

=head2 B<unindex_doc>

 $ph->unindex_doc($docname);

Deletes ('unindexes') a document from the database.

=head2 B<get_docs>

 $ph->get_docs();

Starts the retrieval of all the documents stored in the database.
Each entry will bet fetched using the B<fetch_doc()> method.

=head2 B<fetch_doc>

 ($docname,$mtime)=$ph->fetch_doc();

Retrieves sequentially a I<$docname> and its respective I<$mtime>,
triggered by a call to B<get_docs()>. After the last one, undef is
returned.

=head2 B<clear_cache>

 $ph->clear_cache();

Clears the query cache, if one is available in the current driver.
Each successful search is stored in the query cache to speed up
repeated queries, but loses its value when the database has changed,
so this method must be called after a document (or group of documents)
are (re)indexed.

=head2 B<close>

 $ph->close();

Closes the Penetrator object. No further uses of it can be done.

=head2 B<index_file>

 $ret = $ph->index_file($filename [,$docname]);

Indexes a complete file. I<$filename> must be accesible and readable.
Returns 0 if the file was correctly indexed, -1 if file could not be
read, -2 if this version of the file is already indexed (by using
file's mtime; see B<stat()>) or -3 if the indexing could not be
done (probably due to another process concurrently indexed the
same file). The final document name will be I<$docname> if set,
or I<$filename> otherwise.

=head2 B<test_file>

 $ret = $ph->test_file($filename, $mtime [,$docname]);

This function takes I<$docname> and I<$mtime> from the return
value of B<fetch_doc> and unindexes it if: a) filename no longer
exists or b) filename is newer than the indexed one. I<$docname>
can be specified if it's different from filename. Returns 1 if
the file has been unindexed, or 0 if the file is fresh.

=cut

sub new
{
	my ($class,%args) = @_;
	my ($ph);

	$class .= "::".$args{'type'};

	$ph = new $class(%args);
			      
	$ph->{'log'} ||= sub {};
	$ph->{'normalize'} ||= \&Penetrator::normalize;

	return($ph);	      
}


sub normalize
# normalizes a string
{
	my ($str) = @_;

	# delete HTML tokens, if any
	$str =~ s/<[^>]+>//g;

	# convert some locale chars
	$str =~ tr//AEIOUAEIOUaeiouaeiou/;
	$str =~ tr//AEIOUaeiou/;
	$str =~ tr//NnCc/;

	# stores as lowercase
	$str = lc($str);

	# delete all non words
	$str =~ s/\W//g;

	return($str);
}


sub index_file
{
	my ($ph, $filename, $docname) = @_;
	my ($mtime);

	open F, $filename or return(-1);

	# unless set, document name is the filename
	$docname = $filename unless $docname;

	# takes mtime
	($mtime) = (stat($filename))[9];
	($mtime) = (stat($docname))[9] unless $mtime;

	# start indexing
	$ph->begin_doc_index($docname,$mtime) or return(-2);

	# loop through file
	while(<F>)
	{
		chop;
		$ph->doc_words(split(/\W+/));
	}

	# abort if file could not be indexed
	$ph->end_doc_index() or return(-3);

	return(0);
}


sub test_file
{
	my ($ph, $filename, $mtime, $docname) = @_;
	my ($zap) = 1;

	$docname ||= $filename;

	# if file is the same, keep it;
	# if file is newer or inexistent, delete it
	if(-e $filename)
	{
		my ($mt) = (stat($filename))[9];

		$zap = 0 if $mtime == $mt;
	}

	$ph->unindex_doc($docname) if $zap;

	return($zap);
}


sub DESTROY { $_[0]->close; }


=head1 DRIVER SPECIFIC INFORMATION

=cut

####################################################################
# Drivers
####################################################################

package Penetrator::DBM;

@ISA = ("Penetrator");

=head2 DBM Driver

The DBM driver uses Berkeley_DB style index files. The value for
the 'type' argument to I<new> is 'DBM'.

The driver specific parameters for a new Penetrator object are:

=over 4

=item I<index_file>

This parameter is mandatory, and must point to an readable and
writable file where the index will be stored.

=item I<cache_dir>

Contains the name of the directory where the query results will
be stored. If none is specified, the query cache won't be used
(regardless of the value of the 'disable_query_cache' argument).

=back

=cut

sub new
{
	my ($class,%args) = @_;
	my (%idx,$ph);

	return(undef) unless $args{'index_file'};

	require DB_File;

	bless(\%args,$class);
	$ph = \%args;

	# zap database if requested
	$ph->_zap() if $ph->{'zap'};

	tie %idx, "DB_File", $ph->{'index_file'};

	# create CODDOC, if none exists
	$idx{"-CODDOC"} = 1 unless $idx{"-CODDOC"};

	$ph->{'idx'} = \%idx;

	return($ph);
}


sub _zap
{
	my ($ph) = @_;

	unlink $ph->{'index_file'} if $ph->{'index_file'};
}


sub unindex_doc
{
	my ($ph,$docname) = @_;
	my ($idx,$mt,$coddoc);

	$idx = $ph->{'idx'};

	if(($mt,$coddoc) = ( $idx->{"F$docname"} =~ /^(.+),(.+)$/ ) )
	{
		delete $idx->{"D$coddoc"};
		delete $idx->{"F$docname"};

		return(1);
	}

	return(0);
}


sub begin_doc_index
{
	my ($ph,$docname,$mt) = @_;
	my ($idx);

	$idx = $ph->{'idx'};

	# fails if document exists and has the same mt
	if(my ($smt,$coddoc) = ($idx->{"F$docname"} =~ /^(.*),(.*)$/))
	{
		return if $smt == $mt;

		$ph->unindex_doc($docname);
	}

	# gets a new document code
	$ph->{'coddoc'} = $idx->{"-CODDOC"}++;

	# store
	$idx->{"D".$ph->{'coddoc'}} = "$docname";
	$idx->{"F$docname"} = "$mt,".$ph->{'coddoc'};

	return(1);
}


sub doc_words
{
	my ($ph,@words) = @_;
	my ($idx,$coddoc,$mws);

	$coddoc = $ph->{'coddoc'} or return;

	$idx = $ph->{'idx'};
	$mws = $ph->{'min_word_size'};

	foreach my $word (@words)
	{
		$word = $ph->{'normalize'}->($word);

		next if $word =~ /^$/;
		next if length($word) < $mws;

		# ignore if yet stored
		next if $idx->{"W".$word} =~ /\Q<$coddoc>\E/;

		# add doc to this word
		$idx->{"W".$word} .= "<$coddoc>";
	}
}


sub end_doc_index
{
	my ($ph) = @_;

	delete $ph->{'coddoc'};

	return(1);
}


sub _clean_word
{
	my ($ph,$word) = @_;
	my ($idx,$docids,$nd,@l);

	$idx = $ph->{'idx'};

	$docids = $idx->{"W".$w} or return;
	$docids =~ s/\<//g;

	$nd = "";
	foreach my $i (split(/\>/,$docids))
	{
		$nd .= "<".$i.">" if $idx->{"D$i"};
	}

	# write it
	$idx->{"W".$w} = $nd;
}


sub search
{
	my ($ph,@arg_words) = @_;
	my (@words,@anti_words);
	my ($idx,$word,%selection,$docids);
	my (@res,$cache_entry);

	# split words and anti_words
	foreach my $w (@arg_words)
	{
		if($w =~ /^-/ or $w =~ /^~/)
		{ push(@anti_words, $ph->{'normalize'}->($w)); }
		else
		{ push(@words, $ph->{'normalize'}->($w)); }
	}

	$cache_entry = join(':',@words)."-".join(':',@anti_words);

	# try to find it in the query cache, if defined
	if($ph->{'cache_dir'} and not $ph->{'disable_query_cache'})
	{
		if(open F,"$ph->{'cache_dir'}/$cache_entry")
		{
			@res = <F>;
			close F;

			chop(@res);
			$ph->{'res'} = \@res;
			return;
		}
	}

	# finds the first word
	$word = shift(@words);
	$ph->_clean_word($word);

	$idx = $ph->{'idx'};

	$docids = $idx->{"W".lc($word)};
	return unless $docids;

	# fills the selection hash
	$docids =~ s/\<//g;
	foreach my $i (split(/\>/,$docids))
	{
		$selection{$i} = $idx->{"D$i"} if $idx->{"D$i"};
	}

	# if more than one word, calculate the union set
	foreach $word (@words)
	{
		my (%docids);

		$ph->_clean_word($word);

		$docids = $idx->{"W".lc($word)};
		return unless $docids;

		$docids =~ s/\<//g;

		# hashes (for speed) docids that has this word
		foreach my $i (split(/\>/,$docids))
		{
			$docids{$i} = 1 if $i;
		}

		# deletes any previous selection not
		# in this word docs
		foreach my $i (keys(%selection))
		{
			delete $selection{$i}
				unless $docids{$i};
		}
	}

	# delete documents containing the anti-words
	foreach $word (@anti_words)
	{
		$ph->_clean_word($word);

		$docids = $idx->{"W".lc($word)};
		next unless $docids;

		$docids =~ s/\<//g;

		# deletes documents containing this word
		foreach my $i (split(/\>/,$docids))
		{
			delete $selection{$i} if $selection{$i};
		}
	}

	# stores the resulting query
	@res = values(%selection);
	$ph->{'res'} = \@res;

	# store in the query cache
	if($ph->{'cache_dir'} and not $ph->{'disable_query_cache'})
	{
		mkdir $ph->{'cache_dir'};
		chmod 0777, $ph->{'cache_dir'};

		if(open F,">$ph->{'cache_dir'}/$cache_entry")
		{
			foreach my $l (@res) { print F "$l\n"; }
			close F;

			chmod 0666,"$ph->{'cache_dir'}/$cache_entry";
		}
	}
}


sub fetch
{
	my ($ph) = @_;

	return(shift(@{$ph->{'res'}}));
}


sub get_docs
{
	my ($ph) = @_;

	$ph->{'nextdoc'} = 1;
}


sub fetch_doc
{
	my ($ph) = @_;
	my ($idx,$docname,$mtime,@z);

	return(@z) unless $ph->{'nextdoc'};

	$idx = $ph->{'idx'};

	# take next existing one
	while(!($docname = $idx->{'D'.$ph->{'nextdoc'}}) and
		$ph->{'nextdoc'} < $idx->{'-CODDOC'})
	{
		$ph->{'nextdoc'}++;
	}

	# abort if no more docs
	if($ph->{'nextdoc'} >= $idx->{'-CODDOC'})
	{
		delete $ph->{'nextdoc'};
		return(@z);
	}

	$ph->{'nextdoc'}++;

	# take mtime
	($mtime) = split(/,/,$idx->{"F$docname"});

	return(($docname,$mtime));
}


sub clear_cache
{
	my ($ph) = @_;

	if($ph->{'cache_dir'})
	{
		opendir D,$ph->{'cache_dir'} or return;

		foreach my $i (readdir D)
		{
			next if $i =~ /^\.[\.]?$/;
			unlink "$ph->{'cache_dir'}/$i";
		}

		closedir D;
	}
}


sub close
{
	my ($ph) = @_;

	if($ph->{'idx'})
	{
		untie $ph->{'idx'};
		chmod 0666, $ph->{'index_file'};

		delete $ph->{'idx'};
	}
}


####################################################################
# Drivers
####################################################################

package Penetrator::GDBM;

@ISA = ("Penetrator::DBM");

=head2 GDBM Driver

The GDBM driver is very similar to the DBM driver, but uses
the GNU dbm library instead of the Berkeley_DB one. The value for
the 'type' argument to I<new> is 'GDBM'.

In every other aspect is exactly the same as the DBM driver.

=cut

sub new
{
	my ($class,%args) = @_;
	my (%idx,$ph);

	return(undef) unless $args{'index_file'};

	require GDBM_File;
	import GDBM_File;

	bless(\%args,$class);
	$ph = \%args;

	# zap database if requested
	$ph->_zap() if $ph->{'zap'};

	tie %idx, "GDBM_File", $ph->{'index_file'}, &GDBM_WRCREAT, 0666;

	# create CODDOC, if none exists
	$idx{"-CODDOC"} = 1 unless $idx{"-CODDOC"};

	$ph->{'idx'} = \%idx;

	return($ph);
}



####################################################################
# Drivers
####################################################################

package Penetrator::SDBM;

@ISA = ("Penetrator::DBM");

=head2 SDBM Driver

The SDBM driver is very similar to the DBM driver, but uses
the Perl internal SDBM library instead of the Berkeley_DB one.
This library is reportedly slower and less reliable than the
Berkeley_DB or the GDBM ones, so use it only if those are
not available (SDBM is included with Perl, so it's always there).
The value for the 'type' argument to I<new> is 'SDBM'.

Instead of creating just the specified I<index_file>, two
files will, appending I<.dir> and I<.pag> to their names.

In every other aspect is exactly the same as the DBM driver.

=cut

sub new
{
	my ($class,%args) = @_;
	my (%idx,$ph);

	return(undef) unless $args{'index_file'};

	use Fcntl;
	require SDBM_File;

	bless(\%args,$class);
	$ph = \%args;

	# zap database if requested
	$ph->_zap() if $ph->{'zap'};

	tie %idx, "SDBM_File", $ph->{'index_file'}, O_RDWR|O_CREAT, 0666;

	# create CODDOC, if none exists
	$idx{"-CODDOC"} = 1 unless $idx{"-CODDOC"};

	$ph->{'idx'} = \%idx;

	return($ph);
}


sub _zap
{
	my ($ph) = @_;

	if($ph->{'index_file'})
	{
		unlink $ph->{'index_file'}.".dir";
		unlink $ph->{'index_file'}.".pag";
	}
}



####################################################################
# Drivers
####################################################################

package Penetrator::DBI;

@ISA = ("Penetrator");

=head2 DBI Driver

The DBI driver uses the Perl DataBase Interface to relational
databases. It has been tested with MySQL and PostgreSQL, but hopefully
any other DBI-supported RDBMS could be used.

The driver specific parameters for a new Penetrator object are:

=over 4

=item I<dbi_string>

This argument is mandatory (unless you set I<dbh>, see below) and must
store the DBI connection string to your database. Consult your DBI
driver for details.

=item I<dbi_user>

Database user, if necessary.

=item I<dbi_passwd>

Database user's password, if necessary.

=item I<dbh>

If this parameter is set, it must point to an already open DBI handle.
If used, I<dbi_string>, I<dbi_user> and I<dbi_passwd> will be ignored.
Use this if you already have an open database connection to your Penetrator
database.

=item I<transaction>

If set to nonzero, all session will be protected by a transaction.

=item I<insert_retries>

Number of retries before giving up indexing a document (20 by
default). When more than one process is indexing documents on
the same database, the codification of their words can collide and
the insertion fail; this insertion is retried I<insert_retries>
times before failing and returning error from B<end_doc_index()>.

=item I<delay>

Time to wait before retrying the storage of the words, in seconds.
1 second by default.

=item I<key_code_type>

SQL type for the codes that are primary keys. This must be the
native auto incremental numeric type of the database. On
PostgreSQL is 'serial' (the default) and on MySQL
'integer auto_increment'.

=item I<code_type>

SQL type for the codes. This must be a numeric type compatible
with I<key_code_type>. On PostgreSQL and MySQL is 'integer' (also
the default value).

=item I<doc_type>

SQL type for storing document names. Must be string and indexable.
It's 'text' on PostgreSQL (by default) and 'varchar(100)' (or
similar) on MySQL.

=item I<word_type>

SQL type for storing document words. Must be string and indexable.
It's 'text' on PostgreSQL (by default) and 'varchar(100)' (or
similar) on MySQL.

=item I<query_type>

SQL type for storing queries in the query cache. Must be string
and indexable. It's 'text' on PostgreSQL (by default) and
'varchar(100)' (or similar) on MySQL.

=item I<drop_commands>

Reference to a list of additional drop commands to 'zap' a
database. If you use PostgreSQL and need to re-zap an already
existing Penetrator database, you B<must> set the following
values to delete the automatically created sequences that
are not deleted when the corresponding tables are dropped:

 "drop_commands" => [ "drop sequence docs_coddoc_seq" ,
		      "drop sequence words_codword_seq" ],

=item I<create_commands>

Reference to a list of additional drop commands to 'zap' a
database. Use it if you want to create additional indexes,
for example.

=back

=cut

sub new
{
	my ($class,%ph) = @_;
	my ($dbh);

	require DBI;

	$ph{'dbi_user'} ||= "";
	$ph{'dbi_passwd'} ||= "";
	$ph{'insert_retries'} ||= 20;

	unless(ref($ph{'dbh'}))
	{
		$dbh = DBI->connect($ph{'dbi_string'},$ph{'dbi_user'},$ph{'dbi_passwd'},
			{ RaiseError => 0, PrintError => 0,
			  AutoCommit => !$ph{'transaction'} } )
		or return(undef);

		$ph{'dbh'}=$dbh;
	}

	# default SQL types
	$ph{'key_code_type'} ||= "serial";
	$ph{'code_type'} ||= "integer";
	$ph{'doc_type'} ||= "text";
	$ph{'word_type'} ||= "text";
	$ph{'query_type'} ||= "text";

	$ph{'delay'} = 1 unless exists($ph{'delay'});

	bless(\%ph,$class);

	# zap database if requested
	_zap(\%ph) if $ph{'zap'};

	unless($ph{'disable_tmp_table'})
	{
		# search temporary tables
		$dbh->do("create temporary table stmp (coddoc ".
			$ph{'code_type'}." primary key, cnt integer)");
		$dbh->do("create index stmp_cnt on stmp (cnt)");
	}

	# temporary tables for indexing
	$dbh->do("create temporary table itmp ( word $ph{'word_type'} )");
	$dbh->do("create index itmp_word on itmp (word)");
	$dbh->do("create temporary table itmp2 ( word $ph{'word_type'} )");

	# indexing prepared statements
	$ph{'doc_find'} = $dbh->prepare("select coddoc,mtime from docs where filename=?");
	$ph{'new_doc'} = $dbh->prepare("insert into docs (filename,mtime) values (?,?)");
	$ph{'add_to_itmp2'} = $dbh->prepare("insert into itmp2 (word) select ".
		"distinct itmp.word from itmp left join ".
		"words on words.word=itmp.word where codword is NULL");
	$ph{'add_to_words'} = $dbh->prepare("insert into words (word) ".
		"select word from itmp2");
	$ph{'insert'} = $dbh->prepare("insert into itmp values (?)");
	$ph{'add_to_docwords'} = $dbh->prepare("insert into docwords ".
		"select ?,codword,count(codword) ".
		"from words, itmp where ".
		"words.word=itmp.word group by codword");

	return(\%ph);
}


sub _zap
{
	my ($ph) = @_;
	my ($dbh,$l);

	require DBI;

	$dbh = $ph->{'dbh'};

	# run the drop SQL commands
	$l = $ph->{'drop_commands'};
	foreach my $a (@$l) { $dbh->do($a); }

	$dbh->do("drop table docs");
	$dbh->do("drop table words");
	$dbh->do("drop table docwords");
	$dbh->do("drop table cache");

	$dbh->do("create table docs (".
		 "coddoc $ph->{'key_code_type'} primary key,".
		 "filename $ph->{'doc_type'} unique,".
		 "mtime $ph->{'code_type'})")
		or die "DBI error (" . $DBI::errstr . ")";

	$dbh->do("create table words (".
		 "codword $ph->{'key_code_type'} primary key,".
		 "word $ph->{'word_type'} unique)")
		or die "DBI error (" . $DBI::errstr . ")";

	$dbh->do("create table docwords (".
		 "coddoc $ph->{'code_type'},".
		 "codword $ph->{'code_type'},".
		 "cnt integer)")
		or die "DBI error (" . $DBI::errstr . ")";

	# create indexes for docwords
	$dbh->do("create index docwords_codword on docwords ( codword )");
	$dbh->do("create index docwords_coddoc on docwords ( coddoc )");

	$dbh->do("create table cache (".
		 "filename $ph->{'doc_type'},".
		 "cnt integer,".
		 "timestamp integer,".
		 "query $ph->{'query_type'})")
		or die "penetrator: DBI error (" . $DBI::errstr . ")";

	# create index for cache
	$dbh->do("create index cache_query on cache(query, cnt)");

	# run more possible SQL commands
	$l = $ph->{'create_commands'};
	foreach my $a (@$l) { $dbh->do($a); }

}


sub unindex_doc
{
	my ($ph,$docname) = @_;
	my ($dbh,$coddoc);

	$dbh = $ph->{'dbh'};

	# find coddoc
	$ph->{'doc_find'}->execute($docname);
	($coddoc) = $ph->{'doc_find'}->fetchrow_array();
	$ph->{'doc_find'}->finish();

	return(0) unless $coddoc;

	$dbh->do("delete from docs where coddoc='$coddoc'");
	$dbh->do("delete from docwords where coddoc='$coddoc'");

	return(1);
}


sub begin_doc_index
{
	my ($ph,$docname,$mt) = @_;
	my ($dbh,$coddoc,$smt);

	$dbh = $ph->{'dbh'};

	# find if doc exists
	$ph->{'doc_find'}->execute($docname) or die $DBI::errstr;

	if(($coddoc,$smt) = $ph->{'doc_find'}->fetchrow_array())
	{
		# if file is in the database,
		# do nothing and go on
		if($smt == $mt)
		{
			$ph->{'doc_find'}->finish();
			return(0);
		}

		$ph->delete_doc($docname);
	}

	# create document; if this insert fails,
	# the document could have been inserted by another
	# process meanwhile, so just fail
	$ph->{'new_doc'}->execute($docname,$mt) or return(0);

	# portable way of knowing what coddoc
	# has been assigned to last stored document
	# (by auto-incrementing)
	$ph->{'doc_find'}->execute($docname) or die $DBI::errstr;
	($coddoc) = $ph->{'doc_find'}->fetchrow_array();
	$ph->{'doc_find'}->finish();

	# clean tmp
	$dbh->do("delete from itmp");

	# protect in transaction
	$dbh->{AutoCommit} = 0;

	$ph->{'coddoc'}=$coddoc;

	return(1);
}


sub doc_words
{
	my ($ph,@words) = @_;
	my ($mws);

	$mws = $ph->{'min_word_size'};

	foreach my $word (@words)
	{
		$word = $ph->{'normalize'}->($word);

		next if $word =~ /^$/;
		next if length($word) < $mws;

		$ph->{'insert'}->execute($word) or die $DBI::errstr;
	}
}


sub end_doc_index
{
	my ($ph) = @_;
	my ($dbh,$ir,$coddoc);

	$dbh = $ph->{'dbh'};
	$coddoc = $ph->{'coddoc'};

	# stop transaction
	$dbh->{AutoCommit} = !$ph->{'transaction'};

	# add to words table all new words, retrying several times
	for($ir = $ph->{'insert_retries'};$ir > 0;$ir--)
	{
		# delete temporal itmp2
		$dbh->do("delete from itmp2");

		# transfer to tmp2
		$ph->{'add_to_itmp2'}->execute() or die $DBI::errstr;

		# finally move to words
		last if $ph->{'add_to_words'}->execute();

		# adding to words failed (probably because
		# another process has added the same words
		# and they must be unique), so sleep and retry
		$ph->{'log'}->("Can't add to words (" .
				$DBI::errstr . "). Retry...");

		sleep($ph->{'delay'});
	}

	delete $ph->{'coddoc'};

	if($ir == 0)
	{
		# out of retries; document indexing failed
		$dbh->do("delete from docs where coddoc=$coddoc");
		return(0);
	}

	# add all words to docwords
	$ph->{'add_to_docwords'}->execute($coddoc)
		or die $DBI::errstr;

	return(1);
}


sub search
{
	my ($ph,@arg_words) = @_;
	my (@words,@anti_words);
	my ($dbh,$dsh,$query,$count,@res,$q);

	$dbh = $ph->{'dbh'};
	delete $ph->{'dsh'};

	# split words and anti_words
	foreach my $w (@arg_words)
	{
		if($w =~ /^-/ or $w =~ /^~/)
		{ push(@anti_words, $ph->{'normalize'}->($w)); }
		else
		{ push(@words, $ph->{'normalize'}->($w)); }
	}

	unless($ph->{'disable_query_cache'})
	{
		# try first in the cache
		$q = join(" ",@arg_words);

		$dsh = $dbh->prepare("select count(*) from cache where query=?");
		$dsh->execute($q);
		($count) = $dsh->fetchrow_array();

		if($count)
		{
			# query is in the cache: return it
			$dsh = $dbh->prepare(
			"select filename from cache where query=? order by cnt desc");
			$dsh->execute($q);

			$ph->{'dsh'} = $dsh;

			return;
		}
	}

	# build the first query, that retrieves the word codes
	$query = "select codword from words where ";

	foreach(my ($n) = 0;$n < scalar(@words);$n++)
	{
		$query .= " or " if $n;
		$query .= "word='$words[$n]'";
	}

	$ph->{'query_1'} = $query;

	$dsh = $dbh->prepare($query);
	$dsh->execute();

	# build the second query
	$query = "";
	unless($ph->{'disable_tmp_table'})
	{
		$dbh->do("delete from stmp");
		$query = "insert into stmp ";
	}

	$query .= "select coddoc,avg(cnt)+min(cnt) from docwords where ";
	$count = 0;
	while(my (@l) = $dsh->fetchrow_array)
	{
		$query .= " or " if $count;
		$query .= "codword=" . $l[0];

		$count++;
	}

	# exit if some of the words could not be found
	return if $count != scalar(@words);

	# finish query string
	$query .= " group by coddoc having count(coddoc)=$count";

	$ph->{'query_2'} = $query;

	$dsh = $dbh->prepare($query);
	$dsh->execute();

	unless($ph->{'disable_tmp_table'})
	{
		# process the anti-words
		if(scalar(@anti_words))
		{
			# get word codes
			$query = "select codword from words where ";

			foreach(my ($n) = 0;$n < scalar(@anti_words);$n++)
			{
				$query .= " or " if $n;
				$query .= "word='$anti_words[$n]'";
			}

			$ph->{'query_anti_1'} = $query;

			$dsh = $dbh->prepare($query);
			$dsh->execute();

			# find document codes
			$query = "select coddoc from docwords where ";
			$count = 0;
			while(my (@l) = $dsh->fetchrow_array)
			{
				$query .= " or " if $count;
				$query .= "codword=" . $l[0];
				$count++;
			}

			$ph->{'query_anti_2'} = $query;

			$dsh = $dbh->prepare($query);
			$dsh->execute();

			# generate sequential delete commands
			# using no more than a few 'or' clauses
			for(;;)
			{
				my (@l);

				$query = "delete from stmp where ";
				for($count = 0;$count < 32 and
					@l = $dsh->fetchrow_array;$count++)
				{
					$query .= " or " if $count;
					$query .= "coddoc=" . $l[0];
				}

				$ph->{'query_anti_3'} = $query;

				$dbh->do($query) if $count;

				last unless $count == 32;
			}
		}

		my ($t) = time;

		# insert query into the cache
		$dbh->do("insert into cache select filename,cnt,$t,'$q' ".
			"from docs, stmp where docs.coddoc=stmp.coddoc")
				unless $ph->{'disable_query_cache'};

		$query = "select filename from docs, stmp where ".
		       "docs.coddoc=stmp.coddoc order by cnt desc";
	}
	else
	{
		# build the third and final query
		$query = "select filename from docs where ";
		$count = 0;
		while(my (@l) = $dsh->fetchrow_array)
		{
			$query .= " or " if $count;
			$query .= "coddoc=" . $l[0];

			$count++;
		}
	}

	$ph->{'query_3'} = $query;

	# final query
	$dsh = $dbh->prepare($query);
	$dsh->execute();

	$ph->{'dsh'} = $dsh;
}


sub fetch
{
	my ($ph) = @_;
	my ($dsh);

	$dsh = $ph->{'dsh'} or return(undef);

	return(($dsh->fetchrow_array())[0]);
}


sub get_docs
{
	my ($ph) = @_;
	my ($dbh);

	$dbh = $ph->{'dbh'};

	$ph->{'dsh'} = $dbh->prepare("select filename,mtime from docs");
	$ph->{'dsh'}->execute();
}


sub fetch_doc
{
	my ($ph) = @_;

	return($ph->{'dsh'}->fetchrow_array());
}


sub clear_cache
{
	my ($ph) = @_;
	my ($dbh);

	$dbh = $ph->{'dbh'};

	$dbh->do("delete from cache");
}


sub close
{
	my ($ph) = @_;

	my ($dbh) = $ph->{'dbh'};

	if($dbh and $dbh->isa('DBI'))
	{
		$dbh->commit() if $ph->{'transaction'};

		delete $ph->{'dbh'};
	}
}


sub DESTROY { $_[0]->close; }


####################################################################
# Drivers
####################################################################

package Penetrator::plain;

@ISA = ("Penetrator");

=head2 plain Driver

The plain driver is a dummy driver that doesn't generate any index,
but does sequential searches everytime (as a recursive grep). As hoped,
is slow, but no big files are created, so it can be used in disk space
challenged environments, as web hosting services and such.

The driver specific parameters for a new Penetrator object are:

=over 4

=item I<index_file>

This parameter is mandatory and must point to a filename where the
'indexed' files are stored. This file is populated by I<begin_doc_index()>
and used by I<search()> to know which files must be scanned.

=back

=cut

sub new
{
	my ($class, %args) = @_;
	my ($ph, %data);

	return(undef) unless $args{'index_file'};

	bless(\%args,$class);
	$ph = \%args;

	# zap database if requested
	if($ph->{'zap'})
	{
		unlink($ph->{'index_file'});
	}
	else
	{
		# reads the file
		if(open F, $ph->{'index_file'})
		{
			while(<F>)
			{
				chomp;
				my ($key,$value) = /^(.*):(.*)$/;

				$data{$key} = $value;
			}

			close F;
		}
	}

	$ph->{'-data'} = \%data;
	$ph->{'-dirty-data'} = 0;

	return($ph);
}


sub unindex_doc
{
	my ($ph,$docname) = @_;

	# deletes the docname
	delete $ph->{'-data'}->{$docname};

	# marks the data as dirty
	$ph->{'-dirty-data'}++;
}


sub begin_doc_index
{
	my ($ph,$docname,$mt) = @_;
	my ($ret) = 0;

	# if $mt is greater, change it
	if($mt > $ph->{'-data'}->{$docname})
	{
		# update mt and mark as dirty
		$ph->{'-data'}->{$docname} = $mt;

		$ph->{'-dirty-data'}++;

		$ret = 1;
	}

	return($ret);
}


sub doc_words
{
	my ($ph,@words) = @_;

	# this function is dummy
}


sub end_doc_index
{
	my ($ph) = @_;

	# this function is dummy
}


sub search
{
	my ($ph,@arg_words) = @_;
	my ($cache_entry);
	my (@words,@anti_words);
	my @res = ();

	# split words and anti_words
	foreach my $w (@arg_words)
	{
		if($w =~ /^-/ or $w =~ /^~/)
		{ push(@anti_words, $ph->{'normalize'}->($w)); }
		else
		{ push(@words, $ph->{'normalize'}->($w)); }
	}

	$cache_entry = join(':',@words)."-".join(':',@anti_words);

	# try to find it in the query cache, if defined
	if($ph->{'cache_dir'} and not $ph->{'disable_query_cache'})
	{
		if(open F,"$ph->{'cache_dir'}/$cache_entry")
		{
			@res = <F>;
			close F;

			chop(@res);
			$ph->{'res'} = \@res;
			return;
		}
	}

	foreach my $f (keys(%{$ph->{'-data'}}))
	{
		my (%w, $aw);

		foreach my $w (@words)
		{
			$w{$w} = 1;
		}

		next unless open F, $f;

		$aw = 0;

		while (my $l = <F>)
		{
			# tests if any antiword appears
			foreach my $w (@anti_words)
			{
				if($l =~ /\b$w\b/i)
				{
					$aw = 1;
					last;
				}
			}

			# don't look any further if antiword found
			last if $aw;

			foreach my $w (keys(%w))
			{
				# if word is found, don't
				# search it more
				delete $w{$w} if $l =~ /\b$w\b/i;
			}

			if(scalar(keys(%w)) == 0)
			{
				push(@res, $f);
				last;
			}
		}

		close F;
	}

	$ph->{'res'} = \@res;

	# store in the query cache
	if($ph->{'cache_dir'} and not $ph->{'disable_query_cache'})
	{
		mkdir $ph->{'cache_dir'};
		chmod 0777, $ph->{'cache_dir'};

		if(open F,">$ph->{'cache_dir'}/$cache_entry")
		{
			foreach my $l (@res) { print F "$l\n"; }
			close F;

			chmod 0666,"$ph->{'cache_dir'}/$cache_entry";
		}
	}
}


sub fetch
{
	my ($ph) = @_;

	return(shift(@{$ph->{'res'}}));
}


sub get_docs
{
	my ($ph) = @_;

	my @l = keys(%{$ph->{'-data'}});
	$ph->{'-docs'} = \@l;
}


sub fetch_doc
{
	my ($ph) = @_;
	my ($doc,@z);

	return(@z) unless $doc = shift(@{$ph->{'-docs'}});

	@z = ($doc, $ph->{'-data'}->{$doc});

	return(@z);
}


sub clear_cache
{
	my ($ph) = @_;

	if($ph->{'cache_dir'})
	{
		opendir D,$ph->{'cache_dir'} or return;

		foreach my $i (readdir D)
		{
			next if $i =~ /^\.[\.]?$/;
			unlink "$ph->{'cache_dir'}/$i";
		}

		closedir D;
	}
}


sub close
{
	my ($ph) = @_;

	# if data is dirty, flush to disk
	if($ph->{'-dirty-data'})
	{
		if(open F, ">".$ph->{'index_file'})
		{
			foreach my $d (keys(%{$ph->{'-data'}}))
			{
				print F $d.":".$ph->{'-data'}->{$d}."\n";
			}

			close F;
		}

		$ph->{'-dirty-data'} = 0;
	}
}


#######################################################################

=head1 ADVANCED TOPICS

=head2 Word normalization

Penetrator normalizes each word by converting it to lowercase,
extracting non-ASCII characters, deleting HTML tags and substituting
several latin language specific characters (mainly accented vocals) to
its ASCII equivalents. This normalization is done when storing the
words and when performing the search. The default behaviour can be
changed by providing your own normalization function. You can do
this by supplying it to the object initialization function with the
I<'normalize'> option, as in

 $ph = new Penetrator("type" => $driver,
	"normalize" => \&my_normalization_function,
	[ ... other arguments ]
	);

This function must accept one scalar parameter (the word), and return
the same value after normalization. This is probably the simplest example:

 sub my_normalization_function
 # only convert to lowercase
 {
	my ($word) = @_;
	return(lc($word));
 }

But you probably want a more sophisticated one, as this let pass
punctuation characters that will become part of the word (that will,
consequently, not be found when searched without them).

=cut

=head1 AUTHOR

Angel Ortega angel@triptico.com

=cut

1;
